using System;
using System.Collections.Generic;
using System.Linq;
using System.Security.Policy;
using Leap;
using MacGruber;
using SimpleJSON;
using UnityEngine;

namespace CheesyFX
{
    public class FlexRegion : MonoBehaviour
    {
        protected RBDynamics[] rbDynamics;
        public List<FlexRegionSupportTracker> supportTrackers = new List<FlexRegionSupportTracker>();
        public int id;
        public new string name;
        protected Atom atom => FlexIt.singleton.containingAtom;
        public List<List<DAZMorph>> morphs = new List<List<DAZMorph>>();
        protected List<List<MorphBuffer>> morphBuffers = new List<List<MorphBuffer>>();
        protected bool[] doDyn = {true,true,true,true};
        protected bool[] doStatic = {true,true,true,true};
        public bool supported;
        public FlexRegion[] children;
        
        public HeelTrigger lHeelTrigger;
        public HeelTrigger rHeelTrigger;
        public HeelTrigger lToeTrigger;
        public HeelTrigger rToeTrigger;
        
        public int morphsCount => morphs.Count;

        public JSONStorableBool enabledJ = new JSONStorableBool("Enabled", true);
        public JSONStorableVector4 dynamicStrengthJ = new JSONStorableVector4(
            "Dynamic",
            new Vector4(1f, 1f, 1f, 1f),
            new Vector4(),
            new Vector4(10f,10f,10f, 10f));
        public JSONStorableVector4 amountJ = new JSONStorableVector4(
            "Amount",
            new Vector4(2f, 2f, 2f, 2f),
            new Vector4(),
            new Vector4(5f,5f,5f, 5f));
        public JSONStorableVector4 minAccelJ = new JSONStorableVector4(
            "minAcc",
            new Vector4(0f, 0f, 0f, 0f),
            new Vector4(),
            new Vector4(20f,20f,20f, 20f));
        
        public JSONStorableVector4 sRatioJ = new JSONStorableVector4(
            "S-Ratio",
            new Vector4(.75f,.75f,.75f, .75f),
            new Vector4(0f, 0f, 0f, 0f),
            new Vector4(1f,1f,1f, 1f));

        protected Dictionary<int, string> sides = new Dictionary<int, string>
        {
            { 0, "L" }, { 1, "R" }
        };

        public Vector4 sRatio { get; set; }
        public Vector4 dynamicStrength { get; set; }
        public Vector4 staticStrength { get; set; }

        public Vector4 minAccel { get; set; }
        // private Vector4 maxAccel { get; set; }
        protected Vector4 amount { get; set; }

        public virtual void OnEnable()
        {
            supportTrackers.ForEach(x => x.SetActive(true));
        }

        public virtual void Init(int id, string[] rbNames)
        {
            this.id = id;
            GetMorphs();
            
            enabledJ.setCallbackFunction += val => enabled = val;
            dynamicStrengthJ.setCallbackFunction += val => OnDynamicStrengthChanged(val);
            minAccelJ.setCallbackFunction += val => minAccel = val;
            sRatioJ.setCallbackFunction += OnSRatioChanged;
            amountJ.setCallbackFunction += OnAmountChanged;
            SetupDefaults();
            
        }

        protected virtual void GetMorphs()
        {
            for (int i = 0; i < 4; i++)
            {
                bool morphsFound = false;
                List<DAZMorph> sideMorphs = new List<DAZMorph>();
                for (int j = 0; j < 2; j++)
                {
                    string side = sides[j];
                    var morph = FlexIt.GetMorphByUid($"FI_{name}{i}{side}");
                    if (morph == null) continue;
                    morph.jsonFloat.constrained = false;
                    sideMorphs.Add(morph);
                    morphsFound = true;
                }

                if (!morphsFound) continue;
                morphs.Add(sideMorphs);
                morphBuffers.Add(sideMorphs.Select(x => new MorphBuffer(x)).ToList());
            }
        }

        protected virtual void SetupSupportTrackers(Rigidbody rb)
        {
        }
        
        public virtual void UpdateSupportTrackers()
        {
            lToeTrigger = lHeelTrigger = rToeTrigger = rHeelTrigger = null;
            supportTrackers.ForEach(x => x.Destroy());
            supportTrackers.Clear();
            foreach (var rbDynamic in rbDynamics)
            {
                SetupSupportTrackers(rbDynamic.rigidbody);
            }
        }

        public void OnAmountChanged(Vector4 val)
        {
            amount = val;
            OnDynamicStrengthChanged(dynamicStrengthJ.val);
            staticStrength = MultiplyV4(sRatio, amount);
            for (int i = 0; i < morphBuffers.Count; i++)
            {
                for(int j=0; j<morphBuffers[i].Count; j++) morphBuffers[i][j].max = amount[i];
            }
        }

        public void OnSRatioChanged(Vector4 val)
        {
            sRatio = val;
            doDyn = new[] {val[0] < 1f, val[1] < 1f, val[2] < 1f, val[3] < 1f};
            doStatic = new[] {val[0] > 0f, val[1] > 0f, val[2] > 0f, val[3] > 0f};
            staticStrength = MultiplyV4(sRatio, amount);
        }

        public void OnDynamicStrengthChanged(Vector4 val)
        {
            dynamicStrength = MultiplyV4(val, amountJ.val);
        }

        protected virtual void SetupDefaults()
        {
            OnAmountChanged(amountJ.val);
            OnSRatioChanged(sRatioJ.val);
        }

        protected virtual float GetFlex(int direction, int side)
        {
            return rbDynamics[side].rigidbody.transform.forward.y;
        }

        protected virtual bool Flexed(float flex, int direction, int side)
        {
            return direction == 0 ? flex > 0f : flex < 0f;;
        }
        
        protected virtual float GetAcceleration(int direction, int side)
        {
            float val = rbDynamics[side].angularAcceleration.x;
            val *= amount[direction];
            return val;
        }

        public virtual void FixedUpdate()
        {
            float speed = FlexIt.flexRegressionSpeed;
            for (int direction = 0; direction < morphs.Count; direction++)
            {
                if(amount[direction] == 0f) continue;
                bool dyn = doDyn[direction];
                bool stat = doStatic[direction];
                for (int side = 0; side < morphs[direction].Count; side++)
                {
                    morphs[direction][side].max = amount[direction];
                    float mVal = morphs[direction][side].morphValue;
                    float mTarget = 0f;
                    bool flexed = false;
                    float flex = 0f;
                    if(stat)
                    {
                        if((supportTrackers.Count == 0 || !supportTrackers[side].supported) && 
                           (children == null || !children.Any(x => x.supportTrackers[side].supported)))
                        {
                            flex = GetFlex(direction, side);
                            flexed = Flexed(flex, direction, side);
                        }
                    }
                    if(flexed)
                    {
                        flex = Mathf.Abs(flex);
                        flex = ShakeIt.sin[(int)(flex * 90)];
                        mTarget = flex * staticStrength[direction];
                    }
                    if(Mathf.Abs(mVal - mTarget) > .02f)
                    {
                        morphs[direction][side].morphValue = Mathf.Lerp(mVal, mTarget, speed);
                    }
                    if(dyn)
                    {
                        
                        float f = GetAcceleration(direction, side);
                        bool inRange = (direction == 1 || direction == 3) ? f > minAccel[direction] : f < -minAccel[direction];
                        if (inRange)
                        {
                            morphBuffers[direction][side].Fill(Mathf.Abs(f) * dynamicStrength[direction] * FlexIt.realFrametime);
                        }
                        morphBuffers[direction][side].Pop();
                    }
                }
            }
        }

        public virtual void Reset()
        {
            for (int i = 0; i < morphs.Count; i++)
            {
                for(int j=0; j<morphs[i].Count; j++) morphs[i][j].morphValue = 0f;
            }
        }

        public virtual void OnDisable()
        {
            // for (int i = 0; i < rbDynamics.Length; i++) rbDynamics[i].uses--;
            supportTrackers.ForEach(x => x.SetActive(false));
            Reset();
        }

        public virtual void OnDestroy()
        {
            supportTrackers.ForEach(x => x.Destroy());
            foreach (var rbDynamic in rbDynamics)
            {
                Destroy(rbDynamic);
            }
        }

        private Vector4 MultiplyV4(Vector4 a, Vector4 b)
        {
            for (int i = 0; i < 4; i++) a[i] *= b[i];
            return a;
        }
        
        public void Store(JSONClass jc, bool forceStore = false)
        {
            JSONClass tempJC = new JSONClass();
            enabledJ.Store(tempJC, true);
            amountJ.Store(tempJC, true);
            sRatioJ.Store(tempJC, true);
            dynamicStrengthJ.Store(tempJC, true);
            jc[name] = tempJC;
        }

        public void Load(JSONClass jc)
        {
            if (jc.HasKey(name))
            {
                jc = jc[name].AsObject;
                enabledJ.Load(jc);
                amountJ.Load(jc);
                sRatioJ.Load(jc);
                dynamicStrengthJ.Load(jc);
            }
        }
    }

    public class MorphBuffer
    {
        private DAZMorph morph;
        private float buffer;
        public float max;
        public float increment = .25f;

        public MorphBuffer(DAZMorph morph)
        {
            this.morph = morph;
        }

        public void Fill(float val)
        {
            buffer += val;
        }

        public void Pop()
        {
            if (buffer < .001f) return;
            float val = buffer * increment;
            buffer -= val;
            float newVal = morph.morphValue + val;
            morph.morphValue = newVal > max ? max : newVal;
        }
    }
}