using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using Leap.Unity;
using MeshVR;
using MacGruber;
using UnityEngine;
using UnityEngine.Events;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.Serialization;
using Utils = MacGruber.Utils;


namespace CheesyFX {

	public class WorkOut : MVRScript
	{
		private JSONStorableBool adjustOffset = new JSONStorableBool("Adjust Gloss Texture Offset", true);
		public JSONStorableFloat gloss;
		public JSONStorableFloat glossTextureOffset;
        public JSONStorableFloat power = new JSONStorableFloat("Current Power", 0f, 0f, 2f, false, false);
        public JSONStorableFloat workDone = new JSONStorableFloat("Work Done Recently", 0f, 0f, 10000f, false, false);
        public JSONStorableFloat exhaustion = new JSONStorableFloat("Exhaustion", 0f, 0f, 10f, true, false);

		public JSONStorableFloat powerToWork = new JSONStorableFloat("Power To Work Factor", 1f, 0f, 10f, false, true);
		public JSONStorableFloat regenerationRate = new JSONStorableFloat("Regeneration Rate", 2f, 0f, 10f, false, true);
		public JSONStorableFloat sweatBuildUp = new JSONStorableFloat("Sweat Gain", 1f, 0f, 5f, false, true);
		public JSONStorableFloat sweatRegression = new JSONStorableFloat("Sweat Regression", 1f, 0f, 2f, false, true);
		
		public JSONStorableFloat maxGloss = new JSONStorableFloat("Max gloss", 8f, 5f, 8f, true, true);
		public JSONStorableFloat minGloss = new JSONStorableFloat("Min gloss", 5f, 0f, 7.5f, true, true);
		public JSONStorableFloat maxGlossTextureOffset = new JSONStorableFloat("Max Gloss Texture Offset", .6f, -1f, 1f, true, true);
		public JSONStorableFloat minGlossTextureOffset = new JSONStorableFloat("Min Gloss Texture Offset", .45f, -1f, 1f, true, true);
		public JSONStorableFloat glossTextureOffsetThreshold = new JSONStorableFloat("Gloss Texture Offset Threshold", 6.6f, 4.7f, 8f, true, true);
		public JSONStorableFloat exhaustionGain = new JSONStorableFloat("Exhaustion Gain", .5f, 0f, 10f, false, true);
		public JSONStorableFloat maxExhaustion = new JSONStorableFloat("Max Exhaustion", 10f, 0f, 20f, false, true);

		public CustomFloatTrigger currentPowerTrigger;
		public CustomFloatTrigger workDoneTrigger;
		public CustomFloatTrigger exhaustionTrigger;
		public CustomFloatTrigger glossTrigger;

		private DAZMorph[] flexMorphs;
		public JSONStorableString glossInfo = new JSONStorableString("glossInfo", "");
		JSONStorableParam[] tweakParams;
		
		private RBDynamics[] rbDynamics;
		private bool initialized;

		private List<object> UIElements = new List<object>();
		private UnityEventsListener uiListener;

		public override void Init()
		{
			UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements);
			Utils.OnInitUI(CreateUIElement);
			SimpleTriggerHandler.LoadAssets();
			StartCoroutine(DeferredInit());
		}

		private IEnumerator DeferredInit()
		{
			JSONStorable skin = containingAtom.GetStorable("skin", true);
			while (SuperController.singleton.isLoading || (object)skin == null)
			{
				yield return new WaitForEndOfFrame();
				skin = containingAtom.GetStorable("skin", true);
			}
			gloss = skin.GetFloatJSONParam("Gloss");
			// minGloss.val = gloss.val;
			glossTextureOffset = skin.GetFloatJSONParam("Gloss Texture Offset");
			
			currentPowerTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			workDoneTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			exhaustionTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			glossTrigger = gameObject.AddComponent<CustomFloatTrigger>();
			currentPowerTrigger.Init(this, power, "", false, 1f, 0f , 2f);
			workDoneTrigger.Init(this, workDone, "", false, 5000f, 0f , 10000f);
			exhaustionTrigger.Init(this, exhaustion, "", false, 2f, 0f , 10f);
			glossTrigger.Init(this, gloss, "", false, 7f, 5f , 8f);
			
			maxGloss.setCallbackFunction += (x) => SetAlpha();
			minGloss.setCallbackFunction += (x) => SetAlpha();
			maxGlossTextureOffset.setCallbackFunction += (x) => SetAlpha();
			minGlossTextureOffset.setCallbackFunction += (x) => SetAlpha();
			glossTextureOffsetThreshold.setCallbackFunction += (x) => SetAlpha();
			maxExhaustion.setCallbackFunction += val => exhaustion.max = val;
			
			// exhaustionGain.setCallbackFunction += (x) => { if(x == 0f) exhaustion = 0f; };
			JSONStorableAction actionReset = new JSONStorableAction("Reset", () => { Reset(); });
			RegisterAction(actionReset);

			SetAlpha();
			flexMorphs = FileManagerSecure.GetFiles(ShakeIt.flexMorphPath).Where(x => x.EndsWith(".vmi"))
				.Select(x => ShakeIt.morphUI.GetMorphByUid(FileManagerSecure.NormalizePath(x))).ToArray();
				
			rbDynamics = containingAtom.forceReceivers.Select(x => x.gameObject.AddRbDynamic()).ToArray();

			tweakParams = new JSONStorableParam[]{
				powerToWork,
				regenerationRate,
				sweatBuildUp,
				sweatRegression,
				maxGloss,
				minGloss,
				maxGlossTextureOffset,
				minGlossTextureOffset,
				exhaustionGain,
				maxExhaustion,
				glossTextureOffsetThreshold,
			};
			foreach(var tweakParam in tweakParams) tweakParam.Register(this);
			CreateUI();
			
			enabledJSON.val = enabledJSON.defaultVal = false;
			initialized = true;
		}
		
		public override void InitUI()
		{
			base.InitUI();
			if(UITransform == null || uiListener != null) return;
			uiListener = UITransform.gameObject.AddComponent<UnityEventsListener>();
			uiListener.onEnabled.AddListener(() => UIManager.SetScript(this, CreateUIElement, leftUIElements, rightUIElements));
		}

		private void OnDestroy()
		{
			Destroy(uiListener);
		}

		protected void Reset()
		{
			exhaustion.val   = 0f;
			workDone.val = 0f;
			gloss.val    = minGloss.val;
		}

		public void FixedUpdate()
		{
			if (!SuperController.singleton.freezeAnimation)
			{
				power.val = GetActivity()*Time.fixedDeltaTime;
				if (power.val > .1f)
				{
					workDone.val += power.val * powerToWork.val;
					if(exhaustionGain.val > 0f){
						if(workDone.val > 1000f) exhaustion.val = Mathf.Lerp(exhaustion.val, maxExhaustion.val, Time.fixedDeltaTime * exhaustionGain.val * workDone.val * 10e-7f);
						
					}
				}
				else if (exhaustion.val > 0f)
				{
					exhaustion.val = Mathf.Lerp(exhaustion.val, 0f, .005f*Time.fixedDeltaTime);
				} 

				if (workDone.val > 0f)
				{
					float diff = maxGloss.val - gloss.val;
					gloss.val += 10e-9f * sweatBuildUp.val * (exhaustion.val)* workDone.val * diff * diff;
					workDone.val = Mathf.Lerp(workDone.val, 0f, .01f*regenerationRate.val*Time.fixedDeltaTime);
					gloss.val = Mathf.Lerp(gloss.val, minGloss.val, .01f*sweatRegression.val*Time.fixedDeltaTime);
				}
				
				
				// if(gloss.val > glossTextureOffsetThreshold.val){
				// 	glossTextureOffset.val = minGlossTextureOffset.val + alpha * (maxGloss.val - gloss.val)*(maxGloss.val - gloss.val);
				// }
				// else if(glossTextureOffset.val != maxGlossTextureOffset.val) glossTextureOffset.val = maxGlossTextureOffset.val;
				if (adjustOffset.val)
				{
					if (gloss.val <= glossTextureOffsetThreshold.val)
					{
						float diff = minGloss.val - gloss.val;
						glossTextureOffset.val = maxGlossTextureOffset.val + alpha2 * diff * diff * diff;
					}
					else if (glossTextureOffset.val != minGlossTextureOffset.val)
						glossTextureOffset.val = minGlossTextureOffset.val;
				}

				if(UITransform.gameObject.activeSelf) 
					glossInfo.val = $"Gloss: {gloss.val:0.000}  Offset: {glossTextureOffset.val:0.000}  Exh: {exhaustion:0.00}";
			}
		}

		private float alpha;
		private float alpha2;
		private void SetAlpha()
		{
			glossTextureOffsetThreshold.max = maxGloss.val;
			glossTextureOffsetThreshold.min = minGloss.val;
			// alpha = (maxGlossTextureOffset.val - minGlossTextureOffset.val) / 
			//         ((maxGloss.val - glossTextureOffsetThreshold.val) * (maxGloss.val - glossTextureOffsetThreshold.val));

			float diff = minGloss.val - glossTextureOffsetThreshold.val;
			alpha2 = (minGlossTextureOffset.val - maxGlossTextureOffset.val) / (diff*diff*diff);
		}
		
		public float GetActivity()
		{
			float act = 0f;
			for (int i = 0; i < rbDynamics.Length; i++)
			{
				act += rbDynamics[i].activity;
			}

			for (int i = 0; i < flexMorphs.Length; i++)
			{
				var mVal = flexMorphs[i].morphValue;
				if (mVal > .05f) act += mVal * 2f / flexMorphs[i].max;
			}
			// act.Print();
			return act;
		}

		public void ClearUI()
		{
			UIManager.RemoveUIElements(UIElements);
		}
		
		public void CreateUI()
		{
			try
			{
				UIDynamicButton button;
				UIDynamicSlider slider;
				UIDynamicToggle toggle;
				UIDynamicTextInfo infoline;
				JSONStorableString info;

				// CreateToggle(enabledJSON)
				adjustOffset.CreateUI(UIElements:UIElements, script:this);
				float height = 50f;
				infoline = Utils.SetupInfoOneLine(this, glossInfo, true);
				infoline.height = height;
				UIElements.Add(infoline);

				power.CreateUI(UIElements:UIElements, script:this);
				workDone.CreateUI(UIElements:UIElements, script:this,rightSide:true);
				foreach(var jsonParam in tweakParams)
				{
					jsonParam.CreateUI(UIElements:UIElements, script:this, rightSide:Array.IndexOf(tweakParams, jsonParam)%2 == 1);
				}

				// action triggers
				button = Utils.SetupButton(this, "Current Power Triggers", () =>
				{
					ClearUI();
					currentPowerTrigger.OpenPanel(this, CreateUI);
				}, true);
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				UIElements.Add(button);
				button = Utils.SetupButton(this, "Work Done Triggers", () =>
				{
					ClearUI();
					workDoneTrigger.OpenPanel(this, CreateUI);
				}, true);
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				UIElements.Add(button);
				button = Utils.SetupButton(this, "Exhaustion Triggers", () =>
				{
					ClearUI();
					exhaustionTrigger.OpenPanel(this, CreateUI);
				}, true);
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				UIElements.Add(button);
				button = Utils.SetupButton(this, "Gloss Triggers", () =>
				{
					ClearUI();
					glossTrigger.OpenPanel(this, CreateUI);
				}, true);
				button.buttonColor = new Color(0.45f, 1f, 0.45f);
				UIElements.Add(button);
			}
			catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        public JSONClass MyGetJSON(JSONClass jc){
            // JSONClass jc = new JSONClass();
			maxGloss.Store(jc);
			minGloss.Store(jc);
			maxGlossTextureOffset.Store(jc);
			minGlossTextureOffset.Store(jc);
			glossTextureOffsetThreshold.Store(jc);
			powerToWork.Store(jc);
			maxExhaustion.Store(jc);
			regenerationRate.Store(jc);
			sweatBuildUp.Store(jc);
			sweatRegression.Store(jc);
			
            return jc;
        }

        public void Load(JSONClass jc){
	        enabledJSON.Load(jc);
			maxGloss.Load(jc);
			minGloss.Load(jc);
			maxGlossTextureOffset.Load(jc);
			minGlossTextureOffset.Load(jc);
			glossTextureOffsetThreshold.Load(jc);
			powerToWork.Load(jc);
			maxExhaustion.Load(jc);
			regenerationRate.Load(jc);
			sweatBuildUp.Load(jc);
			sweatRegression.Load(jc);
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
	        JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
	        jc = MyGetJSON(jc);
			// store action triggers
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[currentPowerTrigger.name] = currentPowerTrigger.Store(subScenePrefix);
				jc[workDoneTrigger.name] = workDoneTrigger.Store(subScenePrefix);
				jc[exhaustionTrigger.name] = exhaustionTrigger.Store(  base.subScenePrefix);
				jc[glossTrigger.name] = glossTrigger.Store(subScenePrefix);
			}
			return jc;
        }

        public override void LateRestoreFromJSON(
	        JSONClass jc,
	        bool restorePhysical,
	        bool restoreAppearance,
	        bool setMissingToDefault
        )
        {
	        StartCoroutine(DeferredLateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault));
        }

        private IEnumerator DeferredLateRestoreFromJSON(
	        JSONClass jc,
	        bool restorePhysical,
	        bool restoreAppearance,
	        bool setMissingToDefault)
        {
	        while (!initialized) yield return new WaitForEndOfFrame();
	        base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
	        Load(jc);
			// restore action triggers
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				currentPowerTrigger.Load(jc, subScenePrefix);
				workDoneTrigger.Load(jc, subScenePrefix);
				exhaustionTrigger.Load(jc, subScenePrefix);
				glossTrigger.Load(jc, subScenePrefix);
			}
        }
	}
}